from dataclasses import dataclass
import logging
from typing import Any, List, Optional

from aioinflux3.exceptions import LogicException


@dataclass
class Field:
    key: str
    val: Any

    def serialize(self, force_noquot=False):
        if force_noquot:
            return f'{self.key}={self.val}'
        elif isinstance(self.val, str):
            return f'{self.key}="{self.val}"'
        else:
            return f'{self.key}={self.val}'

@dataclass
class Measurement:
    name: str
    tag: Optional[List[Field]]
    fields: Optional[List[Field]]
    ts: int
    
    @classmethod
    def new(cls, name, ts, tag: List[Field] = None, fields: List[Field] = None):
        return cls(name=name, tag=tag, fields=fields, ts=ts)
    
    def serialize(self):
        tag_str = ",".join([field.serialize(force_noquot=True)
                           for field in self.tag]) if self.tag else ""
        filed_set = ",".join([field.serialize() for field in self.fields])
        if tag_str:
            return f'{self.name},{tag_str} {filed_set} {self.ts}\n'
        else:
            return f'{self.name} {filed_set} {self.ts}\n'


@dataclass
class QueryBody:
    params: dict
    query: str

    def serialize(self):
        return {
            'params': self.params,
            'query': self.query
        }


class Query:
    def __init__(self, bucket, measurement=''):
        self.lines = []
        self.measurement = measurement
        self.lines.append(f'from(bucket: "{bucket}")')
        self.apply_range = False
        self.apply_filter = False
        self.apply_window = False
        self.apply_yield = False
        self.params = {}

    def range(self, start: str = '', end: str ='') -> 'Query':
        # self.params['start'] = start
        # self.params['end'] = end
        if not end:
            self.lines.append(f"  |> range(start: {start} )")
        else:
            self.lines.append(f"  |> range(start: {start}, stop: {end})")
        self.lines.append(
            '  |> filter(fn: (r)=> r["_measurement"] == "tokens")')
        self.apply_range = True
        return self
    
    def filter(self, col: str, **conditions) -> 'Query':
        if not self.apply_range:
            raise LogicException('Must apply range first')
        self.lines.append(f'  |> filter(fn: (r) => r["_field"] == "{col}")')
        if conditions:
            for k, v in conditions.items():
                self.lines.append(f'  |> filter(fn: (r) => r["{k}"] == "{v}")')
        self.apply_filter = True
        return self

    def window(self, every='', fn='', createEmpty=False) -> 'Query':
        if not self.apply_range:
            raise LogicException('Must apply range first')
        if not self.apply_filter:
            raise LogicException('Must apply filter first')
        self.lines.append(
            f'  |> aggregateWindow(every: {every}, fn: {fn}, createEmpty: {"true" if createEmpty else "false"})')
        self.apply_window = True
        return self
    
    def do_yield(self, name='') -> 'Query':
        if not self.apply_range:
            raise LogicException('Must apply range first')
        if not self.apply_filter:
            raise LogicException('Must apply filter first')
        if name:
            self.lines.append(f'  |> yield(name: "{name}")')
        else:
            self.lines.append(f'  |> yield()')
        self.apply_yield = True
        return self

    def to_query(self) -> str:
        if not self.apply_range:
            raise LogicException('Must apply range first')
        if not self.apply_filter:
            raise LogicException('Must apply filter first')
        if not self.apply_yield:
            raise LogicException('Must yield')
        query = "\n".join(self.lines)
        return query


    


    
        
        




from .client import InfluxClient
from .line_protocal import Measurement, QueryBody, Field, Query
from .exceptions import LogicException


__all__ = ['InfluxClient', 'Measurement', 'Field', 'Query', 'LogicException']



import logging
import httpx
import urllib
import time
import datetime
import numpy as np
from typing import List
from aioinflux3.line_protocal import Measurement, QueryBody, Query

UTC_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
UTF_SHORT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

class InfluxClient:
    def __init__(self, host:str = '127.0.0.1', port: int = 8086, token: str = '', bucket: str = '', org: str = ''):
        self.url_base = f'http://{host}:{port}/api/v2'
        self.headers = {
            "Accept": "application/json",
            "Authorization": f"Token {token}"
        }
        self.bucket = bucket
        self.org = org

    def query_string(self):
        return urllib.parse.urlencode({
            'bucket': self.bucket,
            'org': self.org,
            'precision': 'ms'
        })

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass

    def url_for(self, path: str) -> str:
        return f"{self.url_base}{path}?{self.query_string()}"

    async def write(self, data: Measurement):
        post_url = self.url_for('/write')
        logging.error('will post to:%s', post_url)
        async with httpx.AsyncClient() as client:
            resp = await client.post(post_url, content=data.serialize(), headers=self.headers)
            logging.error(resp.text)

    def table_response(self, lines: List[str]) -> dict:
        header = lines.pop(0).split(',')[2:]
        tbody = []
        for row in lines:
            cols = [col.strip() for col in row.split(',')[2:]]
            tbody.append(cols)
        return dict(header=header, tbody=tbody)

    def _trans_value(self, val, idx, dtype):
        tp = dtype[idx]
        if tp == 'int64':
            return int(val)
        if tp == 'float64' or tp == 'float32':
            return float(val)
        return val

    def numpy_response(self, lines: List[str]) -> np.ndarray:
        lines.pop(0)
        columns = {
            'ts':[],
        }
        columns_type = {}
        dtype = [('ts', 'float32')]
        for row in lines:
            cols = [col.strip() for col in row.split(',')]
            if len(cols)<2:
                continue
            _tb = cols[2]
            _dt = datetime.datetime.strptime(cols[5], UTC_FORMAT) if len(
                cols[5]) > 20 else datetime.datetime.strptime(cols[5], UTF_SHORT_FORMAT)
            _ts = time.mktime(_dt.timetuple()) + float("0.%s" % _dt.microsecond)
            _v = cols[6]
            _f = cols[7]
            columns['ts'].append(_ts)
            if _f in columns:
                columns[_f].append(_v)
            else:
                tp = columns_type.get(_tb)
                if not tp:
                    # 没有检测类型，开始检测类型
                    if _v.isdigit():
                        tp = 'int64'
                    elif len([1 for _n in _v.split('.') if _n.isdigit()]) == 2:
                        tp = 'float64'
                    else:
                        tp = 'S16'
                    columns_type[_tb] = tp
                if  _f in columns:
                    columns[_f].append(_v)
                else:
                    dtype.append((_f, tp))
                    columns[_f] = [_v, ]
        columns['ts'] = columns['ts'][:len(columns[_f])]
        data_columns = []
        for c_name, c_tp in dtype:
            data_columns.append(tuple(columns[c_name]))
        arr = np.rot90(np.array(data_columns))
        return np.array([tuple([self._trans_value(col, _idx, dtype) for _idx, col in enumerate(row)]) for row in arr.tolist()], dtype=dtype)


    async def query(self, query: Query, numpy=False):
        post_url = self.url_for('/query')
        body = QueryBody(params={}, query=query.to_query())
        async with httpx.AsyncClient() as client:
            resp = await client.post(post_url, json=body.serialize(), headers=self.headers)
            lines = resp.text.split('\r\n')
            if numpy:
                return self.numpy_response(lines)
            else:
                return self.table_response(lines)



from distutils.core import setup
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='aioinflux3',
      version='0.0.4',
      description='influxdb python client working with asyncio',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Alexander.Li',
      author_email='superpowerlee@gmail.com',
      url='https://github.com/ipconfiger/aio-influx',
      install_requires=[
          "httpx",
          "numpy"
      ],
      packages=['aioinflux3', ],
)
